import os
import sys
import struct
import xml.etree.ElementTree as xml
import numpy as np
import blueprints as blue
from collections import defaultdict
from itertools import count
from imageio import imread



class BaseCache(blue.UniqueThing, blue.CacheType):

	"""
	:class:`Caches <BaseCache>` are used to store large amounts of data. For blueprints to allow for the 
	modification of this data, it has to be loaded to memory. If this data is part of a Thing which is copied 
	multiple times, this would drastically increase the memory usage for redundant data. Caches solve this, 
	by implementing a single source of data for multiple copies of a Thing. Thing classes that contain large 
	amounts of data can simply outsource them to a cache (by linking the outsourced properties to the caches 
	properties) while keeping small attributes that are modified frequently. To ensure data integrity, if a 
	Things property which is stored in its :class:`Cache <BaseCache>` is modified, a new copy of the caches 
	content is created replacing the old ``cache`` attribute of the Thing (unless only one Thing uses the 
	Cache, in which case the Cache is modified directly). For example, consider the following scenario:

	.. code-block:: python
		:caption: Cache example — The large data attribute is outsourced to a Cache

		>>> one = SomeThing(data=open('large.file').read())
		>>> one.cache
		SomeCache<None:2>
		>>> one.data is one.cache.data
		True
		>>> two = one.copy()
		>>> one.cache is two.cache
		True
		>>> one.name = 'one'
		>>> two.name = 'two'
		>>> one.cache is two.cache
		True
		>>> two.data = open('other.file').read()
		>>> one.cache is two.cache
		False
	"""

	def __init__(self, **kwargs):
		"""
		Parameters
		----------
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self._DEPENDENCY_FLAGS = defaultdict(lambda: True)
		super().__init__(**kwargs)


	def _flag_dependencies(self, attr):
		"""
		This method updates the flags for deriving attributes, notifying them that their value mus be 
		updated.
		
		Parameters
		----------
		attr : str
			 name of the attribute
		"""
		if hasattr(self, '_DEPENDENCIES') and hasattr(self, '_DEPENDENCY_FLAGS'):
			for dependency in self._DEPENDENCIES[attr]:
				self._DEPENDENCY_FLAGS[dependency] = False


	def _validate(self, attr):
		"""
		Indicates whether the attribute has to be updated.
		
		Parameters
		----------
		attr : str
			 name of the attribute
		"""
		return not hasattr(self, '_DEPENDENCY_FLAGS') or self._DEPENDENCY_FLAGS[attr]



class MeshCache(blue.MeshCacheType, BaseCache):

	"""
	This class stores the mesh data which might be loaded from a file and is too large to be copied frequently.
	
	Attributes
	----------
	built : bool
		A flag indicating whether the MeshAssets has been built. This is necessary since multiple 
		:class:`MeshAssets <blueprints.assets.MeshAsset>` might use the same :class:`MeshCache` in 
		which case it should be built only once.
	centered : bool
		If ``True`` the vertecies are normalized such that their mean position is the reference frames 
		origin.
	vertecies : list
		The list of vertecies. Each vertex is a np.ndarray with 3 components for each of the spatial 
		dimensions.
	texcoords : list
		The list of texture coordinates.
	texcoords_idx : dict
		A dictionary used to link texture coordinates with face vertecies.
	normals : list
		The normals used by mujoco, this attribute is derived and should not be used for modification, 
		see :attr:`face_normals` or :attr:`vertex_normals` for this instead.
	face_normals : list
		Face normals are the normal vectors of faces. They specify the direction in which the face 
		points.
	vertex_normals : list
		Vertex normals are the normal vectors of vertecies. They specify the direction in which the 
		vertecies of a face point. Specifying vertex normals instead of face normals enables mujoco to
		renderer soft edges. 
	normals_idx : TYPE
		A dictionary used to link normals with faces.
	faces : list
		This attribute is used to write to xml. It returns a flattened list of indecies that is 
		written raw to xml. For a structured list of faces see :attr:`faces`.
	filename : str
		The user specified file name.
	"""
	
	@blue.restrict
	def __init__(self, 
		     vertecies: np.ndarray|list[np.ndarray|list[float|int]]|None = None,  
		     faces:     list[np.ndarray|list[float|int]]|None            = None,
		     filename:  str|None = None, 
		     centered:  bool     = False, 
		     **kwargs) -> None:
		"""
		Parameters
		----------
		vertecies : np.ndarray | list[np.ndarray | list[float | int]] | None, optional
			A list of all vertex positions.
		faces : list[np.ndarray | list[float | int]] | None, optional
			A list of all faces. A face consists of at least 3 indecies of vertecies.
		file : str | None, optional
			The filename from which the mesh data is loaded and to to which the mesh will be saved. 
			The file is saved in a special directory such that no input file is overwritten by the 
			output file.
		centered : bool, optional
			If ``True`` the vertecies are normalized such that their mean position is the reference 
			frames origin.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.centered = centered
		self.filename = filename
		self.texcoords      = None
		self.texcoords_idx  = None
		self.normals        = None
		self.normals_idx    = None
		self.face_normals   = None
		self.vertex_normals = None
		self._DEPENDENCIES  = {'vertecies': ('vertecies_minimum', 
						     'vertecies_center', 
						     'vertecies_maximum')}
		super().__init__(**kwargs)
		if filename is not None:
			self.load(filename)
		elif vertecies is not None:
			self.vertecies = vertecies
			self.faces     = faces
		self._built = False
		if centered and hasattr(self, '_vertecies'):
			self.vertecies = self.vertecies - self.vertecies_center
		#blue.REGISTER.caches.append(self)
		if 'name' not in kwargs:
			self._name = None


	@blue.restrict
	def copy(self, **kwargs) -> blue.CacheType:
		"""
		This method constructs a copy of the Cache with possible alterations to its 
		attributes as specified in kwargs.
		
		Parameters
		----------
		**kwargs
			Keyword arguments are passed to the :meth:`__init__` to replace the attributes 
			of the current Thing from which copy is called.
		
		Returns
		-------
		blue.CacheType
			A new instance of the Cache
		"""
		mesh = super().copy(**kwargs)
		if 'name' not in kwargs:
			mesh._name = self._name
		return mesh


	@blue.restrict
	def _build(self, dirname: str, **kwargs) -> None:
		"""
		This method is called by the Assets parent to construct the xml representations of 
		the kinematic tree.
		
		Parameters
		----------
		dirname : str
			The directory name in which the mesh file is saved.
		**kwargs
			Dummy argument
		"""
		if not self._built:
			pathname, basename = os.path.split(self.filename)
			if not pathname.endswith(dirname):
				path = f'{dirname}/{self.ID}_{basename}'
			else:
				path = self.filename
			self.save(path)
			self._path = path
			self._built = True


	@blue.restrict
	@classmethod
	def _from_xml_element(cls, xml_element) -> blue.CacheType:
		"""
		This method reconstructs an MeshCache from an xml element.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a MeshCache is reconstructed.
		
		Returns
		-------
		blue.CacheType
			The reconstructed MeshCache.
		"""
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		if 'vertex' in post_args:
			init_args['vertecies'] = post_args['vertex']
			del post_args['vertex']
		obj = object.__new__(cls)
		obj.__init__(**init_args)
		for key, val in post_args.items():
			setattr(obj, key, val)
		return obj


	@blue.restrict
	def load(self, filename: str) -> None:
		"""
		This methods loads the mesh data from a file.
		
		Parameters
		----------
		filename : str
			Possible file types are ``'.stl'`` binary or ascii and ``'.obj'`` ascii.
		"""
		with open(filename, 'rb') as file:
			data = file.read()
		if data.isascii():
			self._load_ascii(filename, data.decode('ascii'))
		else:
			self._load_binary(filename, data)


	@blue.restrict
	def _load_binary(self, 
			 filename: str, 
			 data: bytes) -> None:
		"""
		Helper function that routes the loading for difference file formats in binary.
		
		Parameters
		----------
		filename : str
			Possible file types are ``'.stl'``.
		data : bytes
			The data of the file to be parsed
		
		Raises
		------
		NotImplemented
			If the file format is not supported an error is raised.
		"""
		if filename.lower().endswith('.obj'):
			raise NotImplemented
		elif filename.lower().endswith('.stl'):
			self._load_STL_binary(data)
		else:
			raise NotImplemented

	@blue.restrict
	def _load_ascii(self, 
			filename: str, 
			data: str) -> None:
		"""
		Helper function that routes the loading for difference file formats in ascii.
		
		Parameters
		----------
		filename : str
			Possible file types are ``'.stl'`` and ``'.obj'``.
		data : str
			The data of the file to be parsed
		
		Raises
		------
		NotImplemented
			If the file format is not supported an error is raised.
		"""
		#return
		if filename.lower().endswith('.obj'):
			self._load_OBJ_ascii(data)
		elif filename.lower().endswith('.stl'):
			self._load_STL_ascii(data)
		else:
			raise NotImplemented


	@blue.restrict
	def _load_OBJ_ascii(self, data: str) -> None:
		"""
		Parses the data from an ascii obj file to the Cache.
		
		Parameters
		----------
		data : str
			The data of the file to be parsed
		"""
		# DATA CONTAINERS
		vertecies      = []
		faces          = []
		texcoords      = []
		face_normals   = []
		vertex_normals = []
		# SORTING CONTAINERS
		all_face_normals = {}
		texcoords_idx    = {}
		normals_idx      = {}
		# READING LINES
		for line in data.split('\n'):
			# REMOVING COMMENTS
			line = line.split('#')[0]
			# HANDLING VERTECIES
			if line.startswith('v '):
				values = line[2:].strip().split(' ')[:3]
				vertecies.append(tuple(map(float, values)))
			elif line.startswith('vn '):
				values = line[3:].strip().split(' ')[:3]
				vertex_normals.append(np.array(values, dtype=np.float64))
			elif line.startswith('vt '):
				values = line[3:].strip().split(' ')[:2]
				texcoords.append(tuple(map(float, values)))
			elif line.startswith('f '):
				vertex_idx   = line[2:].strip().split(' ')#[:3]
				triangle_idx = [[vertex_idx[0], a, b] for a, b in zip(vertex_idx[1:], vertex_idx[2:])]
				for triangle in triangle_idx:
					if all(map(lambda x: x.count('//') == 1, triangle)):
						# GET VALUES
						values, normal_idx = zip(*(value.split('//') for value in triangle))
						values     = list(map(lambda x: int(x) - 1, values))
						normal_idx = list(map(lambda x: int(x) - 1, normal_idx))
						# SET INDECIES
						normals_idx[len(faces)] = normal_idx
					elif all(map(lambda x: x.count('/') == 2, triangle)):
						# GET VALUES
						values, tex_idx, normal_idx = zip(*(value.split('/') for value in triangle))
						values     = list(map(lambda x: int(x) - 1, values))
						tex_idx    = list(map(lambda x: int(x) - 1, tex_idx))
						normal_idx = list(map(lambda x: int(x) - 1, normal_idx))
						# SET INDECIES
						normals_idx[len(faces)] = normal_idx
						texcoords_idx[len(faces)]  = tex_idx
					elif all(map(lambda x: x.count('/') == 1, triangle)):
						# GET VALUES
						values, tex_idx = zip(*(value.split('/') for value in triangle))
						values  = list(map(lambda x: int(x) - 1, values))
						tex_idx = list(map(lambda x: int(x) - 1, tex_idx))
						# SET INDECIES
						texcoords_idx[len(faces)] = tex_idx
					face = list(map(int, values))
					face = [idx if idx != -2 else len(vertecies) for idx in face]
					faces.append(face)
		assert not texcoords_idx.values() or all(map(texcoords_idx.__contains__, range(len(faces))))
		self.vertecies      = vertecies
		self.faces          = faces
		self.vertex_normals = vertex_normals or None
		self.texcoords      = texcoords      or None
		self.texcoords_idx  = texcoords_idx  or None
		self.normals_idx    = normals_idx    or None


	@blue.restrict
	def _load_STL_ascii(self, data: str) -> None:
		"""
		Parses the data from an ascii stl file to the Cache.
		
		Parameters
		----------
		data : str
			The data of the file to be parsed
		"""
		# CROP DATA
		data = data[data.find('\n'):].split('facet')[1::2]
		# INIT CONTAINERS
		faces          = list()
		normals        = list()
		vertecies      = list()
		vertex_counter = count()
		vertex_table   = defaultdict(lambda: next(vertex_counter))
		# VECTOR CONSTRUCTION FUNCTION
		vector_tuple = lambda s: tuple(map(float, list(filter(lambda x: x != '', s.split(' ')))[:3])) 
		for facet in data:
			# GET FACET SECTION
			lines = facet.split('\n')
			# GET VECTORS
			normal_line    = lines[0].replace('normal', '').strip()
			vertex_0_line  = lines[2].replace('vertex', '').strip()
			vertex_1_line  = lines[3].replace('vertex', '').strip()
			vertex_2_line  = lines[4].replace('vertex', '').strip()
			# CONSTRUCT VECTOR TUPLES
			normal_tuple   = vector_tuple(normal_line)
			vertex_tuple_0 = vector_tuple(vertex_0_line)
			vertex_tuple_1 = vector_tuple(vertex_1_line)
			vertex_tuple_2 = vector_tuple(vertex_2_line)
			# CONSTRUCT NP.ARRAYS
			normal_array   = np.array(normal_tuple)
			vertex_array_0 = np.array(vertex_tuple_0)
			vertex_array_1 = np.array(vertex_tuple_1)
			vertex_array_2 = np.array(vertex_tuple_2)
			# ADD NEW VERTECIES TO HASH TABLE
			if vertex_tuple_0 not in vertex_table:
				vertex_table[vertex_tuple_0]
				vertecies.append(vertex_array_0)
			if vertex_tuple_1 not in vertex_table:
				vertex_table[vertex_tuple_1]
				vertecies.append(vertex_array_1)
			if vertex_tuple_2 not in vertex_table:
				vertex_table[vertex_tuple_2]
				vertecies.append(vertex_array_2)
			# GET NORMAL FROM ORDER
			edge_array_0 = vertex_array_1 - vertex_array_0
			edge_array_1 = vertex_array_2 - vertex_array_0
			edge_cross   = np.cross(edge_array_0, edge_array_1)
			edge_normal  = edge_cross/np.linalg.norm(edge_cross)
			correlation = np.dot(edge_normal, normal_array)
			if correlation > 0:
				vertex_tuples = [vertex_tuple_0, vertex_tuple_1, vertex_tuple_2]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(normal_array)
			elif correlation < 0:
				vertex_tuples = [vertex_tuple_0, vertex_tuple_2, vertex_tuple_1]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(-normal_array)
			else:
				vertex_tuples = [vertex_tuple_0, vertex_tuple_2, vertex_tuple_1]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(normal_array)
				vertex_tuples = [vertex_tuple_0, vertex_tuple_1, vertex_tuple_2]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(-normal_array)
		# SET ATTRIBUTES
		self.vertecies    = vertecies
		self.faces        = faces
		self.face_normals = normals


	@blue.restrict
	def _load_STL_binary(self, data: bytes) -> None:
		"""
		Parses the data from an binary stl file to the Cache.
		
		Parameters
		----------
		data : bytes
			The data of the file to be parsed
		"""
		# CONVERSION FUNCTIONS
		bytes_to_float = lambda x: struct.unpack('f', x)[0]
		vector_tuple   = lambda x: tuple(map(bytes_to_float, (x[:4], x[4:8], x[8:])))
		# HEADER DATA
		header         = data[:80]
		number         = struct.unpack('I', data[80:84])[0]
		# TRTIANGLE DATA
		triangle_data  = data[84:]
		faces          = list()
		normals        = list()
		vertecies      = list()
		vertex_counter = count()
		vertex_table   = defaultdict(lambda: next(vertex_counter))
		for i in range(number):
			# GET TRIANGLE
			triangle = triangle_data[i*50:(i+1)*50]
			# CONSTRUCT VECTOR TUPLES
			normal_tuple   = vector_tuple(triangle[:12])
			vertex_tuple_0 = vector_tuple(triangle[12:24])
			vertex_tuple_1 = vector_tuple(triangle[24:36])
			vertex_tuple_2 = vector_tuple(triangle[36:48])
			# CONSTRUCT NP.ARRAYS
			normal_array   = np.array(normal_tuple)
			vertex_array_0 = np.array(vertex_tuple_0)
			vertex_array_1 = np.array(vertex_tuple_1)
			vertex_array_2 = np.array(vertex_tuple_2)
			# ADD NEW VERTECIES TO HASH TABLE
			if vertex_tuple_0 not in vertex_table:
				vertex_table[vertex_tuple_0]
				vertecies.append(vertex_array_0)
			if vertex_tuple_1 not in vertex_table:
				vertex_table[vertex_tuple_1]
				vertecies.append(vertex_array_1)
			if vertex_tuple_2 not in vertex_table:
				vertex_table[vertex_tuple_2]
				vertecies.append(vertex_array_2)
			# GET NORMAL FROM ORDER
			edge_array_0 = vertex_array_1 - vertex_array_0
			edge_array_1 = vertex_array_2 - vertex_array_0
			edge_cross   = np.cross(edge_array_0, edge_array_1)
			edge_normal  = edge_cross/np.linalg.norm(edge_cross)
			correlation = np.dot(edge_normal, normal_array)
			if correlation > 0:
				vertex_tuples = [vertex_tuple_0, vertex_tuple_1, vertex_tuple_2]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(normal_array)
			elif correlation < 0:
				vertex_tuples = [vertex_tuple_0, vertex_tuple_2, vertex_tuple_1]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(-normal_array)
			else:
				vertex_tuples = [vertex_tuple_0, vertex_tuple_2, vertex_tuple_1]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(normal_array)
				vertex_tuples = [vertex_tuple_0, vertex_tuple_1, vertex_tuple_2]
				face = list(map(vertex_table.__getitem__, vertex_tuples))
				faces.append(face)
				normals.append(-normal_array)
		# SET ATTRIBUTES
		self.vertecies    = vertecies
		self.faces        = faces
		self.face_normals = normals


	@blue.restrict
	def save(self, file: str) -> None:
		"""
		This method saves the Mesh data to a file.
		
		Parameters
		----------
		file : str
			The name to which the file is saved.
		
		Raises
		------
		NotImplemented
			If the file format is not supported an error is raised.
		"""
		if file.lower().endswith('.obj'):
			self._save_OBJ(file)
		elif file.lower().endswith('.stl'):
			self._save_STL(file) 
		elif file.lower().endswith('.msh'):
			raise NotImplemented
		else:
			raise NotImplemented


	@blue.restrict
	def _save_OBJ(self, filename: str|None) -> None:
		"""
		This method saves the Mesh data to an ascii obj file.
		
		Parameters
		----------
		filename : str | None
			The name to which the file is saved.
		"""
		lines = ['# Exported with microcosm AI blueprints\n', '\n# VERTECIES\n']
		for v in self.vertecies:
			lines.append(f'v {v[0]:.6f} {v[1]:.6f} {v[2]:.6f}\n')
		if self.vertex_normals:
			lines.append('\n# NORMALS\n')
			for n in self.vertex_normals:
				lines.append(f'vn {n[0]:.6f} {n[1]:.6f} {n[2]:.6f}\n')
		if self.texcoords:
			lines.append('\n# TEXTURE COORDINATES\n')
			for t in self.texcoords:
				lines.append(f'vt {t[0]:.6f} {t[1]:.6f}\n')
		if self.faces:
			texcoords_idx = self.texcoords_idx
			normals_idx = self.normals_idx
			lines.append('\n# FACES\n')
			for i, face in enumerate(self.faces):
				has_tex = i in texcoords_idx
				has_nrm = i in normals_idx
				if has_tex and has_nrm:
					parts = ' '.join(f"{v + 1}/{t + 1}/{n + 1}" for v, t, n in zip(face, texcoords_idx[i], normals_idx[i]))
				elif has_tex:
					parts = ' '.join(f"{v + 1}/{t + 1}" for v, t in zip(face, texcoords_idx[i]))
				elif has_nrm:
					parts = ' '.join(f"{v + 1}//{n + 1}" for v, n in zip(face, normals_idx[i]))
				else:
					parts = ' '.join(f"{v + 1}" for v in face)
				lines.append(f'f {parts}\n')
		with open(filename, 'w') as file:
			file.writelines(lines)


	@blue.restrict
	def _save_STL(self, filename: str) -> None:
		"""
		This method saves the Mesh data to an binary stl file.
		
		Parameters
		----------
		filename : str
			The name to which the file is saved.
		"""
		# GET VERTECIES AND CONSTRUCT CONVERSION FUNCTION
		vertecies      = self.vertecies
		float_to_bytes = lambda x: struct.pack('f', x)
		vector_bytes   = lambda x: float_to_bytes(x[0]) + float_to_bytes(x[1]) + float_to_bytes(x[2])			
		# CREATE HEADER
		SOURCE = bytes('Saved from Blueprints, UNITS= m', encoding='ascii')
		HEADER = bytearray(84)
		HEADER[:len(SOURCE)] = SOURCE
		HEADER[80:] = struct.pack('I', len(self.faces))
		with open(filename,  'wb') as file:
			file.write(HEADER)
			for normal, indecies in zip(self.face_normals, self.faces):
				vertex_a, vertex_b, vertex_c = map(vertecies.__getitem__, indecies)
				file.write(vector_bytes(normal))
				file.write(vector_bytes(vertex_a))
				file.write(vector_bytes(vertex_b))
				file.write(vector_bytes(vertex_c))
				file.write(bytes.fromhex('00') * 2)

	# MUJOCO PROPERTIES

	@property
	def file(self) -> str | None:
		"""
		Filename for Mujoco XML

		Returns
		-------
		str
		"""
		return self.filename


	@property
	def vertex(self):
		"""
		Returns
		-------
		np.ndarray
			This attribute is used to write construct the mujoco xml. It returns a flattened list of 
			coordinates that is written raw to xml. For a structured list of vertecies see :attr:`vertecies`.
		"""
		return self._vertecies.flatten()


	@property
	def face(self) -> np.ndarray|None:
		"""
		Returns
		-------
		np.ndarray | None
			This attribute is used to write construct the mujoco xml. It returns a flattened list of 
			coordinates that is written raw to xml. For a structured list of faces see :attr:`faces`.
		"""
		return self._faces


	@face.setter
	@blue.restrict
	def face(self, face: list) -> None:
		"""
		This attribute is used to write construct the mujoco xml. It returns a flattened list of 
		coordinates that is written raw to xml. For a structured list of faces see :attr:`faces`.
		
		Parameters
		----------
		face : list
			A flattened list of all face indecies.
		"""
		self.faces = face


	@property
	def texcoord(self) -> np.ndarray|None:
		"""
		Returns
		-------
		np.ndarray | None
			This attribute is used to write to xml. It returns a flattened list of texture coordinates that 
			is written raw to xml. For a structured list of texture coordinates see :attr:`texcoords`.
		"""
		return self._texcoords or None


	@texcoord.setter
	@blue.restrict
	def texcoord(self, texcoord: list) -> None:
		"""
		This attribute is used to write to xml. It returns a flattened list of texture coordinates that 
		is written raw to xml. For a structured list of texture coordinates see :attr:`texcoords`.
		
		Parameters
		----------
		texcoord : list
			A flattened list of all texture coordinates.
		"""
		self.texcoords = texcoord


	@property
	def normal(self) -> np.ndarray|None:
		"""
		Returns
		-------
		np.ndarray | None
			This attribute is used to write to xml. It returns a flattened list of face normals that 
			is written raw to xml. For a structured list of face normals see :attr:`normals`.
		"""
		if self._face_normals is None:
			return None
		else:
			return np.concatenate(self._face_normals)


	@normal.setter
	@blue.restrict
	def normal(self, normal: list) -> None:
		"""
		This attribute is used to write to xml. It returns a flattened list of face normals that is 
		written raw to xml. For a structured list of face normals see :attr:`normals`.
		
		Parameters
		----------
		normal : list
			A flattened list of all normals coordinates.
		"""
		self.face_normals = normal


	@property
	def size(self) -> np.ndarray:
		"""
		Returns
		-------
		np.ndarray
			The size of the Mesh is computed as the difference of the minimal and maximal component 
			for each axis on each vertex.
		"""
		return self.vertecies_max - self.vertecies_min
		

	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]) -> None:
		"""
		The size of the Mesh is defined as the difference of the minimal and maximal component for each 
		axis on each vertex.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			Setting this attribute results in a rescaling of all vertecies. Setting this attribute 
			should only be done if the raw data must be modified directly, use :attr:`scale <blueprints.assets.MeshAsset.scale>`
			instead.
		"""
		vertecies = np.array(self.vertecies)
		vertecies = vertecies - self.vertecies_center
		scale     = np.array(size) / self.size
		scale[np.isinf(scale)] = 0
		vertecies = vertecies * scale
		self.vertecies = vertecies + self.vertecies_center

	# BLUEPRINTS PROPERTIES

	@property
	def vertecies(self) -> np.ndarray:
		"""
		Returns
		-------
		np.ndarray
			The list of vertecies. Each vertex is a np.ndarray with 3 components for each of the 
			spatial dimensions.
		"""
		return self._vertecies


	@vertecies.setter
	@blue.restrict
	def vertecies(self, vertecies: np.ndarray|list[np.ndarray|list[float|int]]) -> None:
		"""
		Parameters
		----------
		vertecies : np.ndarray | list[np.ndarray | list[float | int]]
			The list of vertecies. Each vertex is a np.ndarray with 3 components for each of the 
			spatial dimensions.
		"""
		vertecies = np.array(vertecies)
		self._vertecies = vertecies
		self._flag_dependencies('vertecies')
		self._built = False



	@property
	def faces(self) -> list[np.ndarray]|None:
		"""
		Returns
		-------
		list[np.ndarray] | None
			The list of faces. Each face is a np.ndarray with n indecies for each of the vertecies 
			of the face.
		"""
		return self._faces
		####
		if self._faces is None:
			return None
		else:
			return list(np.rollaxis(self._faces.reshape((-1, 3)), axis=0))


	@faces.setter
	@blue.restrict
	def faces(self, faces: np.ndarray|list[np.ndarray|list[float|int]]|None) -> None:
		"""
		Parameters
		----------
		faces : np.ndarray | list[np.ndarray | list[float | int]] | None
			The list of faces. Each face is a np.ndarray with n indecies for each of the vertecies 
			of the face.
		"""
		if isinstance(faces, np.ndarray):
			faces = list(np.rollaxis(faces, axis=0))
		self._faces = faces
		self._built = False
		return
		###
		if faces is not None:
			faces = np.array(faces, dtype=np.int32).flatten()
		self._faces = faces
		self._built = False


	@property
	def texcoords(self) -> list[np.ndarray]|None:
		"""
		Returns
		-------
		list[np.ndarray] | None
			The list of texture coordinates. Each texture coordinate is a np.ndarray with 
			coordinates in the range [0, 1] specifying the point on the texture which is mapped to
			the referencing face vertex.
		"""
		if self._texcoords is None:
			return None
		else:
			return list(np.rollaxis(self._texcoords.reshape((-1, 2)), axis=0))


	@texcoords.setter
	@blue.restrict
	def texcoords(self, texcoords: np.ndarray|list[np.ndarray|list[float|int]]|None) -> None:
		"""
		Parameters
		----------
		texcoords : np.ndarray | list[np.ndarray | list[float | int]] | None
			The list of texture coordinates. Each texture coordinate is a np.ndarray with 
			coordinates in the range [0, 1] specifying the point on the texture which is mapped to
			the referencing face vertex.
		"""
		if texcoords is not None:
			texcoords = np.array(texcoords, dtype=np.float32).flatten()
		self._texcoords = texcoords
		self._built = False


	@property
	def texcoords_idx(self) -> dict|None:
		"""
		Returns
		-------
		dict | None
			The list of texture coordinate indecies for faces.
		"""
		return self._texcoords_idx


	@texcoords_idx.setter
	@blue.restrict
	def texcoords_idx(self, texcoords_idx: dict|None) -> None:
		"""
		Parameters
		----------
		texcoords_idx : dict | None
			The list of texture coordinate indecies for faces.
		"""
		self._texcoords_idx = texcoords_idx
		self._built = False


	@property
	def face_normals(self) -> list[np.ndarray]|None:
		"""
		Returns
		-------
		list[np.ndarray] | None
			Face normals are the normal vectors of faces. They specify the direction in which the 
			face points.
		"""
		if self._face_normals is None:
			if self._faces is not None:
				faces = self.vertecies[self.faces,:]
				edges = faces[:,:,1:] - faces[:,:,:1]
				cross = np.cross(edges[...,0], edges[...,1], axisa=1, axisb=1)
				norms = cross / np.linalg.norm(cross, axis=1)[:,None]
				self.face_normals = norms
				return self.face_normals
			else:
				return None
		else:
			return self._face_normals


	@face_normals.setter
	@blue.restrict
	def face_normals(self, face_normals: np.ndarray|list[np.ndarray|list[float|int]]|None) -> None:
		"""
		Parameters
		----------
		face_normals : np.ndarray | list[np.ndarray | list[float | int]] | None
			Face normals are the normal vectors of faces. They specify the direction in which the 
			face points.
		"""
		if isinstance(face_normals, np.ndarray):
			face_normals = list(np.rollaxis(face_normals, axis=0))
		self._face_normals = face_normals
		self._built = False


	@property
	def vertex_normals(self) -> list[np.ndarray]|None:
		"""
		Returns
		-------
		list[np.ndarray] | None
			Vertex normals are the normal vectors of vertecies. They specify the direction in which 
			the vertecies of a face point. Specifying vertex normals instead of face normals enables 
			mujoco torenderer soft edges. 
		"""

		if self._vertex_normals is None:
			return None
		else:
			return self._vertex_normals


	@vertex_normals.setter
	@blue.restrict
	def vertex_normals(self, vertex_normals: np.ndarray|list[np.ndarray|list[float|int]]|None) -> None:
		"""
		Parameters
		----------
		vertex_normals : np.ndarray | list[np.ndarray | list[float | int]] | None
			Vertex normals are the normal vectors of vertecies. They specify the direction in which 
			the vertecies of a face point. Specifying vertex normals instead of face normals enables 
			mujoco torenderer soft edges. 
		"""
		if isinstance(vertex_normals, np.ndarray):
			vertex_normals = list(np.rollaxis(vertex_normals, axis=0))
		self._vertex_normals = vertex_normals
		self._built = False


	@property
	def normals_idx(self) -> dict|None:
		"""
		The list of normals indecies for faces.
		
		Returns
		-------
		dict | None
		"""
		return self._normals_idx


	@normals_idx.setter
	@blue.restrict
	def normals_idx(self, normals_idx: dict|None) -> None:
		"""
		Parameters
		----------
		normals_idx : dict | None
			The list of normals indecies for faces.
		"""
		self._normals_idx = normals_idx
		self._built = False


	@property
	def vertecies_min(self):
		"""
		The minimum for each axis component of all vertecies.
		
		Returns
		-------
		np.ndarray
		"""
		if self._validate('vertecies_min'):
			self._vertecies_min = np.min(self.vertecies, axis=0)
			self._DEPENDENCY_FLAGS['vertecies_min'] = True
		return self._vertecies_min


	@property
	def vertecies_max(self):
		"""
		The maximum for each axis component of all vertecies.
		
		Returns
		-------
		np.ndarray
		"""
		if self._validate('vertecies_max'):
			self._vertecies_max = np.max(self.vertecies, axis=0)
			self._DEPENDENCY_FLAGS['vertecies_max'] = True
		return self._vertecies_max


	@property
	def vertecies_center(self):
		"""
		The middle point for each axis component of all vertecies.
		
		Returns
		-------
		np.ndarray
		"""
		return (self.vertecies_min + self.vertecies_max)/2



class HFieldCache(blue.HFieldCacheType, BaseCache):
	def __init__(self, 
		     terrain:  np.ndarray|list[np.ndarray|list[float|int]]|None = None,  
		     filename: str|None = None, 
		     **kwargs):
		self._built = False
		self._world = None
		super().__init__(**kwargs)
		self.filename = filename
		if filename is not None:
			self.load(filename)
		elif terrain is not None:
			self.terrain = terrain
		if 'name' not in kwargs:
			self._name = None


	def __getitem__(self, 
			key: tuple[slice|int]) -> np.ndarray|np.float32:
		"""
		Returns
		-------
		np.ndarray | np.float32
			The values from the selected index/slice of the height field.
		"""
		return self.terrain[key]


	def __setitem__(self, 
			key:   tuple[slice|int], 
			value: int|float|list[int|float]|np.ndarray) -> None:
		"""
		Parameters
		----------
		key : tuple[slice | int]
			The indecies/slices of acces
		value : int|float|list[int|float]|np.ndarray
			The value to be assigned in the selected parts of the field.
		"""
		self.terrain[key] = np.array(value, dtype=np.float32)


	@blue.restrict
	def copy(self, **kwargs) -> blue.CacheType:
		"""
		This method constructs a copy of the Cache with possible alterations to its 
		attributes as specified in kwargs.
		
		Parameters
		----------
		**kwargs
			Keyword arguments are passed to the :meth:`__init__` to replace the attributes 
			of the current Thing from which copy is called.
		
		Returns
		-------
		blue.CacheType
			A new instance of the Cache
		"""
		mesh = super().copy(**kwargs)
		if 'name' not in kwargs:
			mesh._name = self._name
		return mesh


	@blue.restrict
	def _build(self, dirname: str, **kwargs) -> None:
		"""
		This method is called by the Assets parent to construct the xml representations of 
		the kinematic tree.
		
		Parameters
		----------
		dirname : str
			The directory name in which the hfield file is saved.
		**kwargs
			Dummy argument
		"""
		if not self._built:
			filename = self.filename or f'hfield.hf'
			pathname, basename = os.path.split(filename)
			if not pathname.endswith(dirname):
				path = f'{dirname}/{self.ID}_{basename}'
			else:
				path = filename
			self.save(path)
			self._path = path
			self._built = True


	@blue.restrict
	@classmethod
	def _from_xml_element(cls, xml_element) -> blue.CacheType:
		"""
		This method reconstructs an HFieldCache from an xml element.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a HFieldCache is reconstructed.
		
		Returns
		-------
		blue.CacheType
			The reconstructed HFieldCache.
		"""
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		if 'terrain' in post_args:
			init_args['terrain'] = post_args['terrain']
			del post_args['terrain']
		if 'file' in rest_args:
			init_args['filename'] = rest_args['file']
		obj = object.__new__(cls)
		obj.__init__(**init_args)
		for key, val in post_args.items():
			setattr(obj, key, val)
		return obj


	@blue.restrict
	def load(self, filename: str) -> None:
		"""
		This methods loads the mesh data from a file.
		
		Parameters
		----------
		filename : str
			Possible file types are ``'.stl'`` binary or ascii and ``'.obj'`` ascii.
		"""
		if filename.lower().endswith('.hf'):
			with open(filename, 'rb') as file:
				data = file.read()
			self._load_HF(data)
		elif filename.lower().endswith('.png'):
			self._load_PNG(filename)
		else:
			raise Exception('HFields are only implemented for PNG and HF.')


	@blue.restrict
	def _load_HF(self, data: bytes) -> None:
		"""
		Parses the data from an binary stl file to the Cache.
		
		Parameters
		----------
		data : bytes
			The data of the file to be parsed
		"""
		# HEADER DATA
		nrows          = struct.unpack('I', data[0:4])[0]
		ncols          = struct.unpack('I', data[4:8])[0]
		# HEIGHT DATA (batch unpack all floats at once)
		n_floats       = nrows * ncols
		heights        = struct.unpack(f'{n_floats}f', data[8:8 + n_floats * 4])
		# SET ATTRIBUTES
		self.terrain   = np.array(heights, dtype=np.float32).reshape((nrows, ncols))
		

	@blue.restrict
	def _load_PNG(self, filename: str) -> None:
		image        = imread(filename)
		height       = np.mean(image[:,:,:3], axis=2)
		self.terrain = height
	

	@blue.restrict
	def save(self, filename: str) -> None:
		"""
		This method saves the Mesh data to a file.
		
		Parameters
		----------
		file : str
			The name to which the file is saved.
		
		Raises
		------
		NotImplemented
			If the file format is not supported an error is raised.
		"""
		if filename.lower().endswith('.hf'):
			self._save_HF(filename)
		else:
			raise NotImplemented


	@blue.restrict
	def _save_HF(self, filename: str) -> None:
		"""
		This method saves the Mesh data to an binary stl file.
		
		Parameters
		----------
		filename : str
			The name to which the file is saved.
		"""
		# GET VERTECIES AND CONSTRUCT CONVERSION FUNCTION
		int_to_bytes   = lambda x: struct.pack('I', x)
		float_to_bytes = lambda x: struct.pack('f', x)
		with open(filename, 'wb') as file:
			file.write(int_to_bytes(self.nrow))
			file.write(int_to_bytes(self.ncol))
			for height in self.terrain.reshape(-1):
				file.write(float_to_bytes(height))
	
	# DERIVED PROPERTIES

	@property
	def elevation(self):
		"""
		Derived dummy property
		"""
		return self.terrain.flatten()

	@property
	def nrow(self) -> int:
		"""
		Number of rows in the terrain.

		Returns
		-------
		int
		"""
		return int(self.terrain.shape[0])

	@property
	def ncol(self) -> int:
		"""
		Number of collumns in the terrain.

		Returns
		-------
		int
		"""
		return int(self.terrain.shape[1])

	# BLUEPRINTS PROPERTIES

	@property
	def filename(self) -> str:
		"""
		The filename of the hfield data.

		Returns
		-------
		str
		"""
		return self._filename


	@filename.setter
	@blue.restrict
	def filename(self, filename: str|None) -> None:
		self._filename = filename
	

	@property
	def terrain(self) -> np.ndarray:
		"""
		Returns
		-------
		np.ndarray
			The list of vertecies. Each vertex is a np.ndarray with 3 components for each of the 
			spatial dimensions.
		"""
		return self._terrain


	@terrain.setter
	@blue.restrict
	def terrain(self, terrain: np.ndarray|list[np.ndarray|list[float|int]]) -> None:
		"""
		Parameters
		----------
		vertecies : np.ndarray | list[np.ndarray | list[float | int]]
			The list of vertecies. Each vertex is a np.ndarray with 3 components for each of the 
			spatial dimensions.
		"""
		self._terrain = np.array(terrain, dtype=np.float32)
		if self._world is not None:
			mj_hfield = self._world._mj_data.model.hfield(self._index)
			mj_hfield.data = self._terrain
			if self._world._viewer is not None:
				self._world._viewer.update_hfield(self._index)
			#self._world._mj_model.update_hfield(self._index)


	@property
	def file(self) -> str:
		"""
		Dummy variable for the default names of hfield data files.

		Returns
		-------
		str
		"""
		if self._file is None:
			return f'hfield_{self.name}.hf'
		else:
			return self._file
	
	
		





CACHE_THINGS = {'mesh':   MeshCache, 
		'hfield': HFieldCache}


"""
Construction and Attachement
----------------------------

Lattices are a blueprints native Thing. They are not part of the mujoco xml but rather 
a convenience object to help generate copies of a :class:`MoveableThing <blueprints.thing.moveable.MoveableThing>` 
arranged in a grid. Attaching a Lattice to any :class:`NodeThing <blueprints.thing.node.NodeThing>` results 
in its components being attached instead. A Lattice can also be constructed from another Lattice


.. code-block:: python
	:caption: Lattice Attachment:

	>>> world   = blue.World()
	>>> box     = blue.geoms.Box(z=0.5, color='grey')
	>>> lattice = box.lattice(directions=[[2, 0, 0], [0, 1.5, 0]], repetitions=[4, 5])
	>>> world.attach(lattice)
	>>> world.attach(blue.geoms.Plane(color='white'), blue.Light(z=10))


.. code-block:: xml
	:caption: XML Structure:

	<worldbody>
                <light cutoff="360.0" name="anonymous_light" pos="0.0 0.0 10.0" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(0)" pos="0.0 0.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(1)" pos="2.0 0.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(2)" pos="4.0 0.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(3)" pos="6.0 0.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(4)" pos="0.0 1.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(5)" pos="2.0 1.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(6)" pos="4.0 1.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(7)" pos="6.0 1.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(8)" pos="0.0 3.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(9)" pos="2.0 3.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(10)" pos="4.0 3.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(11)" pos="6.0 3.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(12)" pos="0.0 4.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(13)" pos="2.0 4.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(14)" pos="4.0 4.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(15)" pos="6.0 4.5 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(16)" pos="0.0 6.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(17)" pos="2.0 6.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(18)" pos="4.0 6.0 0.5" />
                <geom size="0.5 0.5 0.5" type="box" name="anonymous_box_(19)" pos="6.0 6.0 0.5" />
                <geom size="0.0 0.0 1.0" type="plane" name="anonymous_plane" rgba="1.0 1.0 1.0 1.0" />
        </worldbody>


.. image:: /_static/lattice_example_1.png

This Lattice made from a Lattice of a Box is istelf ofcause just a Lattice of a Box constructed by joining 
the axis of the two Lattices: ``lattice_A`` :math:`\\cong` ``lattice_B``


.. code-block:: python
	:caption: Lattice Creation:

	>>> lattice_B = blue.Lattice(box, directions=[[ 2., 0.0, 0.], 
	>>> 					      [ 0., 1.5, 0.], 
	>>> 					      [10., 0.0, 0.], 
	>>> 					      [ 0., 9.0, 0.]], 
	>>> 				  repetitions=[4, 5, 6, 7])


.. image:: /_static/lattice_example_2.png


Indexing and Attributes
-----------------------
A Lattice can be indexed via tuples of integer and slices. If less axes are indexed than contained by 
the Lattice, the last axes are retrieved fully. Indexed Lattices utilize a :class:`LatticeView <blueprints.utils.view.LatticeView>` 
which handles attribute retrieval and setting.


.. code-block:: python
	:caption: One to One:

	>>> box = blue.geoms.Box(x_length=0.03, 
	>>> 			 y_length=0.01, 
	>>> 			 z_length=0.06, 
	>>> 			 color='grey')
	>>> domino = blue.Body(geoms=box, 
	>>> 		       joints=blue.joints.Free(), 
	>>> 		       z=0.02)
	>>> row = domino.lattice([0, 0.06, 0], [20])
	>>> row[0].alpha = -TAU/32
	>>> row[0].y = 0.0075


.. image:: /_static/domino.gif

In the above example, we start the domino chain by tipping over a single Body indexed by ``row[0]``. 
We can instead also modify multiple ellements with a single value, setting every second domino color to ``'black'`` and ``'white'``.

.. code-block:: python
	:caption: One to Many:

	>>> row[0::2].geoms.color = 'black'
	>>> row[1::2].geoms.color = 'white'

.. image:: /_static/black_white_domino.gif


Lastly we can also assign many values to many elements by providing a list which is mapped one to one to the Lattices components. 
We create a gradient of rainbow colors in the same shape as the Lattice.


.. code-block:: python
	:caption: Many to Many:

	>>> rainbow = blue.gradient('purple', 'red', 'orange', 'yellow', 'green', 'teal', 'blue', 'purple', n_steps=20) 
	>>> row.color = rainbow


.. image:: /_static/rainbow_domino.gif

"""



import numpy as np
import blueprints as blue



class Lattice(blue.LatticeType):
	"""
	A Lattice is constructed from a Thing, a set of directions or axes and the number of repetitions. 
	Lattices are not mujoco objects but blueprints utility.
	"""
	__RESERVED = {'name', 
		      '_directions', 
		      '_repetitions', 
		      '_thing', 
		      '_things'}
	@blue.restrict
	def __init__(self, 
		     thing:       blue.MoveableThingType|blue.LatticeType, 
		     directions:  list[float|int]|np.ndarray|list[list[int|float]]|list[np.ndarray], 
		     repetitions: int|list[int]|np.ndarray, 
		     name:        str|None = None) -> None:
		"""
		Parameters
		----------
		thing : blue.MoveableThingType | blue.LatticeType
			This Thing will consitute the content of the Lattice
		directions : list [ float | int ] | np.ndarray | list [ list [ int | float ] ] | list [ np.ndarray ] 
			A list of axes pointing into the directions of the Lattice.
		repetitions : int | list [ int ] | np.ndarray
			A list of repetitions for each axis with the same length as ``directions``
		name : str | None
			The name of the Lattice (used for displaying purposes).
		"""
		self.name = name if name is not None else 'anonymous_lattice'
		# FORMAT DIRECTIONS
		if isinstance(directions, np.ndarray):
			directions = [directions]
		elif all(isinstance(val, (int, float)) for val in directions):
			directions = [directions]
		if all(isinstance(direction, list) for direction in directions):
			directions = [np.array(direction, dtype=np.float32) for direction in directions]
		self._directions = directions
		# FORMAT REPETITIONS
		if isinstance(repetitions, np.ndarray):
			repetitions = list(map(int, repetitions))
		if isinstance(repetitions, int):
			repetitions = [repetitions]
		self._repetitions = repetitions
		if isinstance(thing, blue.LatticeType):
			# EXTEND AXIS
			self._directions  += thing._directions.copy()
			self._repetitions += thing._repetitions.copy()
			# FORMAT THINGS
			self._thing  = thing._thing.copy()
			self._things = self._create_things(self._thing)
		else:
			# FORMAT THINGS
			self._thing  = thing.copy()
			self._things = self._create_things(self._thing)


	@property
	def n_dim(self) -> int:
		"""
		The number of dimensions in the Lattice
		
		Returns
		-------
		int
		"""
		return len(self._directions)


	def lattice(self, 
		    directions:  list[float|int]|np.ndarray|list[list[int|float]]|list[np.ndarray], 
		    repetitions: int|list[int]|np.ndarray, 
		    name:        str|None = None) -> None:
		"""
		A Lattice created with the parent Lattice as its content Thing (see examples above).
		
		Returns
		-------
		blue.LatticeType
		"""
		
		return blue.Lattice(thing=self, 
				    directions=directions, 
				    repetitions=repetitions, 
				    name=name)


	def __repr__(self) -> str:
		return f'Lattice[{'×'.join(map(str, self._repetitions))}|{self._thing.name}]'


	def _create_things(self, thing):
		def rec(depth, field, direction):
			if depth == 1:
				return [thing.shift(direction) for thing in field]
			else:
				return [rec(depth - 1, subfield, direction) for subfield in field]
		for depth, direction, repetition in zip(range(self.n_dim), self._directions[::-1], self._repetitions[::-1]):
			if depth == 0:
				field = [thing.shift(i * direction) for i in range(repetition)]
			else:
				field = [rec(depth, field, i * direction) for i in range(repetition)]
		return field


	def __iter__(self):
		for item in blue.LatticeView(self._things, self):
			yield item


	def copy(self, **kwargs):
		"""
		Parameters
		----------
		kwargs : dict
			The keyword arguments are passed to the individual component Things copy method.
		
		Returns
		-------
		blue.LatticeType
			A fresh copy of the Lattice
		"""
		
		def rec(depth, field, kwargs):
			if depth == 1:
				return [thing.copy(**kwargs) for thing in field]
			else:
				return [rec(depth - 1, subfield, kwargs) for subfield in field]
		things  = rec(self.n_dim, self._things, kwargs)
		# MIRROR INIT
		lattice = object.__new__(Lattice)
		lattice.name         = self.name
		lattice._directions  = self._directions.copy()
		lattice._repetitions = self._repetitions.copy()
		lattice._thing       = self._thing.copy(**kwargs)
		lattice._things      = things
		return lattice


	def shift(self, 
		  pos: list[int|float]|np.ndarray|None                        = None, 
		  x:   int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		  y:   int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		  z:   int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		  **kwargs):
		"""
		This method creates a copy that is shifted by ``pos``.
		
		Parameters
		----------
		pos : list[int | float] | np.ndarray | blue.MoveableThingType | None
			The amount by which the Thing is shifted, optionally another :class:`MoveableThing` 
			can be passed in which case the position of this Thing is taken.
		globally : bool, optional
			Determining whether the shift is relative to the local orientation or the global orientation.
		x : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the X component of the position is to be changed.
		y : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the Y component of the position is to be changed.
		z : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the Z component of the position is to be changed.
		**kwargs
			Keyword arguments are passed to the :meth:`copy`.
		
		Returns
		-------
		blue.LatticeType
			A shifted copy is returned.
		"""
		if pos is not None and any(arg is not None for arg in [x, y, z]):
			raise ValueError(f'Only argument (pos) or arguments (x, y, z) can be set. Method call got ({', '.join(name for name, val in [('pos', pos), ('x', x), ('y', y), ('z', z)] if val is not None)}).')
		if pos is None:
			x = float(x) if x is not None else 0.
			y = float(y) if y is not None else 0.
			z = float(z) if z is not None else 0.
			pos = np.array([x, y, z], dtype=np.float32)
		def rec(depth, field, pos, kwargs):
			if depth == 1:
				return [thing.shift(pos, **kwargs) for thing in field]
			else:
				return [rec(depth - 1, subfield, pos, kwargs) for subfield in field]
		things  = rec(self.n_dim, self._things, pos, kwargs)
		# MIRROR INIT
		lattice = object.__new__(Lattice)
		lattice._directions   = self._directions.copy()
		lattice._repetitions = self._repetitions.copy()
		lattice._thing       = self._thing.copy(**kwargs)
		lattice._things      = things
		return lattice


	def locate(self, 
		   pos: list[int|float]|np.ndarray|None                        = None, 
		   x:   int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		   y:   int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		   z:   int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		   **kwargs):
		"""
		This method creates a copy that is relocated to ``pos``.
		
		Parameters
		----------
		pos : list[int | float] | np.ndarray | blue.MoveableThingType | None
			The position to which the Thing is shifted, optionally another :class:`MoveableThing` 
			can be passed in which case the position of this Thing is taken.
		globally : bool, optional
			Determining whether the relocation is relative to the local orientation or the global orientation.
		x : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the X component of the position is to be changed.
		y : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the Y component of the position is to be changed.
		z : int | float | np.int32 | np.int64 | np.float32 | np.float64 | None
			This argument can be set if just the Z component of the position is to be changed.
		**kwargs
			Keyword arguments are passed to :meth:`copy`.
		
		Returns
		-------
		blue.LatticeType
			A relocated copy is returned.
		"""
		if pos is not None and any(arg is not None for arg in [x, y, z]):
			raise ValueError(f'Only argument (pos) or arguments (x, y, z) can be set. Method call got ({', '.join(name for name, val in [('pos', pos), ('x', x), ('y', y), ('z', z)] if val is not None)}).')
		if pos is None:
			x = float(x) if x is not None else 0.
			y = float(y) if y is not None else 0.
			z = float(z) if z is not None else 0.
			pos = np.array([x, y, z], dtype=np.float32)
		def rec(depth, field, pos, kwargs):
			if depth == 1:
				return [thing.locate(pos, **kwargs) for thing in field]
			else:
				return [rec(depth - 1, subfield, pos, kwargs) for subfield in field]
		things  = rec(self.n_dim, self._things, pos, kwargs)
		# MIRROR INIT
		lattice = object.__new__(Lattice)
		lattice._directions   = self._directions.copy()
		lattice._repetitions = self._repetitions.copy()
		lattice._thing       = self._thing.copy(**kwargs)
		lattice._things      = things
		return lattice


	@blue.restrict
	def __getitem__(self, keys: int|slice|tuple[int|slice]) -> blue.LatticeViewType|blue.MoveableThingType:
		return blue.LatticeView(self._things, self).__getitem__(keys)



	def __getattr__(self, attr: str):
		if attr in self.__RESERVED:
			return super().__getattr__(attr)
		else:
			return blue.LatticeView(self._things, self).__getattr__(attr)
	

	def __setattr__(self, attr: str, val: object):
		if attr in self.__RESERVED:
			super().__setattr__(attr, val)
		else:
			blue.LatticeView(self._things, self).__setattr__(attr, val)


	#@blue.restrict
	#def __setitem__(self, keys:int|slice|tuple[int|slice], val: list|blue.ThingType|blue.LatticeView):
	#	pass



"""
Placeholders can be attached to a :class:`blueprints.body.Body` to mark prominent locations. They do not appear 
in the resulting xml model and cannot be reconstructed from xml. Is is used to mimic a node in the kinematic 
tree that can be used to specify default attachment locations. If a Thing is attached to a Placeholder it is 
instead attached to the Placeholders parent in a position and orienation as if it would have been attached to 
the Placeholder directly. This can either be used for conveniance if many Things are attached to the same spot 
on the parent Body or in constructing a template with predefined spots for attachment. The following example 
shows how to build the template of a table.

.. code-block::
	:caption: Example table construction

	>>> top = blue.geoms.Box(name='top', x_length=2, y_length=1, z_length=0.002)
	>>> leg = blue.geoms.Box.from_points([0, 0, 0], [0, 0, 1], radius=0.03, name='leg')
	>>> table = blue.Body(name='table', pos=[0, 0, 1], geoms=top)
	>>> table.attach(blue.Placeholder(pos=[ 0.9, 0.4, 0], alpha=PI), 
	>>>              blue.Placeholder(pos=[ 0.9,-0.4, 0], alpha=PI), 
	>>>              blue.Placeholder(pos=[-0.9, 0.4, 0], alpha=PI), 
	>>>              blue.Placeholder(pos=[-0.9,-0.4, 0], alpha=PI))
	>>> table.placeholders.attach(leg)
	>>> table.geoms
	View[5|table.geoms]
	>>> table.attach(blue.Placeholder(name='plate', pos=[-0.7, 0, 0]), 
	>>>              blue.Placeholder(name='plate', pos=[ 0.7, 0, 0]))
	>>> table.placeholders
	['leg_(0)', 'leg_(1)', 'leg_(2)', 'leg_(3)', 'plate_(0)', 'plate_(1)']

.. code-block:: mxml
	:caption: XML structure

	<body pos="0.0 0.0 1.0" name="table">
		<geom size="1.0 0.5 0.001" type="box" name="top" />
		<geom size="0.015 0.015 0.5" type="box" euler="3.1415927 0.0 0.0" pos=" 0.9  0.4 -0.5" name="leg_(0)" />
		<geom size="0.015 0.015 0.5" type="box" euler="3.1415927 0.0 0.0" pos=" 0.9 -0.4 -0.5" name="leg_(1)" />
		<geom size="0.015 0.015 0.5" type="box" euler="3.1415927 0.0 0.0" pos="-0.9  0.4 -0.5" name="leg_(2)" />
		<geom size="0.015 0.015 0.5" type="box" euler="3.1415927 0.0 0.0" pos="-0.9 -0.4 -0.5" name="leg_(3)" />
	</body>
"""
import numpy as np
import blueprints as blue



class Placeholder(blue.PlaceholderType, blue.MoveableThing):

	"""
	This class is available through the shortcut :class:`blueprints.Placeholder <Placeholder>`.

	Placeholders can be attached to a :class:`blueprints.body.Body` to mark special places and orientations 
	for attachment. It will not be included into the xml.

	Parameters
	----------
	pos : list[int | float] | np.ndarray, optional
		Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
	alpha : int | float | None, optional
		(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
	beta : int | float | None, optional
		(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
	gamma : int | float | None, optional
		(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
	name : str | None, optional
		The user specified name of the Body. In the case of a naming conflict the name is appended by an enumeration scheme.
	**kwargs
		Keyword arguments are passed to ``super().__init__``.

	Attributes
	----------
	parent : blue.NodeThingType | None
		When a Thing is attached to the Placeholder, it instead gets attached to the parent and its position 
		and orientation will be adjusted such that it is positioned and oriented as if it were attached of 
		the Placeholder.
	"""
	
	def attach(self, *args):
		"""
		Things that are attached to a Placeholder are instead attached to the Placeholders parent Body 
		in a position and orientation as if they were attached to the Placeholder directly.
		
		Parameters
		----------
		*args
			the Things that sould be attached to the Placeholders :attr:`parent`
		
		Raises
		------
		Expection
			If the Placeholder is not attached to a parent an error is raised.
		"""
		if self.parent is None:
			raise Expection('Placeholders must be attached to a body before a Thing can be attached to them!')
		else:
			args = [arg.rotate(self.alpha, 
					   self.beta, 
					   self.gamma, 
					   center=np.zeros(3), 
					   globally=False).shift(self.pos) for arg in args]
			self.parent.attach(*args)


	def detach(self, *args):
		"""
		This method redirects to the Placeholders parent. This means, that Things that have not been 
		attached to the parent directly (not via the Placeholder) can also be detached with this method.
		
		Parameters
		----------
		*args
			the Things that sould be detached from the Placeholders :attr:`parent`
		"""
		if self.parent is None:
			raise Expection('Placeholders must be attached to a body before a Thing can be detached from them!')
		self.parent.detach(*args)
		


	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		Placeholders are not part of mujoco xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The parent of the Thing
		world : blue.WorldType
			The World from which the initial _meth:`blueprints.world.World.build` method was called.
		"""
		pass



import numpy as np
import blueprints as blue



class Material(blue.MaterialType, blue.NodeThing):
	"""
	Materials can be given to visual Things to specify their reflective properties for the renderer. 
	To apply :class:`Textures <blueprints.texture.BaseTexture>` to Things, they must be attributes 
	of a Material which in turn is assigned to the Thing. Just keeping the default values for 
	Materials and assigning it to :class:`Sites <blueprints.sites.BaseSite>` and 
	:class:`Geoms <blueprints.geoms.BaseGeom>` will have no visual effect.
	
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#asset-material>`__.
	"""
	@blue.restrict
	def __init__(self, 
		     texture:     blue.TextureType|None = None, 
		     texrepeat:   list[int|float]|np.ndarray = [1, 1], 
		     texuniform:  bool      = False, 
		     emission:    int|float = 0.0, 
		     specular:    int|float = 0.5, 
		     shininess:   int|float = 0.5, 
		     reflectance: int|float = 0., 
		     metallic:    int|float =-1.0, 
		     roughness:   int|float =-1.0, 
		     color:       str|object|None = None, 
		     name:        str|None        = None, 
		     asset:       blue.MaterialAssetType|None = None, 
		     copy:        bool = False):
		"""
		Parameters
		----------
		texture : blue.TextureType | None, optional
			A :class:`Texture <blueprints.texture.BaseTexture>` Thing
		texrepeat : list [ int | float ] | np.ndarray, optional
			The number of repetitions of the Texture in two directions
		texuniform : bool, optional
			For :class:`BoxTextures <blueprints.texture.Box>`, this attribute controls how 
			cube mapping is applied.
		emission : int | float, optional
			If set positive, the Material emitts lights.
		specular : int | float, optional
			Specularity in OpenGL. This value should be in the range [0 1].
		shininess : int | float, optional
			Shininess in OpenGL
		reflectance : int | float, optional
			Reflectance is computed without ray-tracing, so only a single reflection level 
			is rendered.
		metallic : int | float, optional
			This attribute is ignored by the default mujoco renderer.
		roughness : int | float, optional
			This attribute is ignored by the default mujoco renderer.
		color : str|object | None, optional
			The color of the Material (if set) will be multiplied with the 
			Things color and (if set) the Materials Texture.
		name : str | None, optional
			Name of the Material
		asset : blue.MaterialAssetType | None, optional
			The :class:`Asset <blueprints.asset.MaterialAsset>` of the Material
		copy : bool, optional
			Dummy argument without effect
		"""
		self._CHILDREN = dict() # Layers might be implemented someday
		if asset is None:
			self._asset = blue.assets.MaterialAsset(texture=texture, 
								texrepeat=texrepeat, 
								texuniform=texuniform, 
								emission=emission, 
								specular=specular, 
								shininess=shininess, 
								reflectance=reflectance, 
								metallic=metallic, 
								roughness=roughness, 
								color=color)
		else:
			self._asset = asset
		self.asset._add(self)
		super().__init__(name=name, color=color)


	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		self._asset._build(parent=parent, 
				   world=world, 
				   indicies=indicies, 
				   **kwargs)
		return self._asset._xml_root


	@property
	def asset(self) -> blue.MaterialAssetType:
		"""
		The :class:`Asset <blueprints.asset.MaterialAsset>` of the Material

		Returns
		-------
		blue.MaterialAssetType
		"""
		return self._asset


	@asset.setter
	@blue.restrict
	def asset(self, asset: blue.MaterialAssetType) -> None:
		#self._asset._remove(self)
		self._asset = asset
		#self._asset._add(self)
	


	@property
	def texrepeat(self) -> np.ndarray:
		"""
		The number of repetitions of the Texture in two directions

		Returns
		-------
		np.ndarray
		"""
		return self._asset.texrepeat


	@texrepeat.setter
	@blue.restrict
	def texrepeat(self, texrepeat: list[int|float]|np.ndarray) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.texrepeat = texrepeat


	@property
	def texuniform(self) -> bool:
		"""
		For :class:`BoxTextures <blueprints.texture.Box>`, this attribute controls how 
		cube mapping is applied. The default value ``False`` means apply cube mapping 
		directly, using the actual size of the object. The value “true” maps the texture 
		to a unit object before scaling it to its actual size (geometric primitives are 
		created by the renderer as unit objects and then scaled). In some cases this 
		leads to more uniform texture appearance, but in general, which settings produces 
		better results depends on the texture and the object. For 2d textures, this attribute 
		interacts with texrepeat above. Let texrepeat be N. The default value “false” means 
		that the 2d texture is repeated N times over the (z-facing side of the) object. 
		The value “true” means that the 2D :class:`PlaneTexture <blueprints.texture.Plane>` 
		is repeated N times over one spatial unit, regardless of object size.

		Returns
		-------
		bool
		"""
		return self.asset.texuniform


	@texuniform.setter
	@blue.restrict
	def texuniform(self, texuniform: bool) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.texuniform = texuniform


	@property
	def emission(self) -> float:
		"""
		If set positive, the Material emitts lights.
		
		Returns
		-------
		float
		"""
		return self.asset.emission


	@emission.setter
	@blue.restrict
	def emission(self, emission: int|float) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.emission = emission


	@property
	def specular(self) -> float:
		"""
		Specularity in OpenGL. This value should be in the range [0 1].

		Returns
		-------
		float
		"""
		return self.asset.specular


	@specular.setter
	@blue.restrict
	def specular(self, specular: int|float) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.specular = specular


	@property
	def shininess(self) -> float:
		"""
		Shininess in OpenGL.

		Returns
		-------
		float
		"""
		return self.asset.shininess


	@shininess.setter
	@blue.restrict
	def shininess(self, shininess: int|float) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.shininess = shininess


	@property
	def reflectance(self) -> float:
		"""
		This attribute should be in the range [0 1]. If the value 
		is greater than 0, and the material is applied to a plane 
		or a box geom, the renderer will simulate reflectance. The 
		larger the value, the stronger the reflectance. For boxes, 
		only the face in the direction of the local +Z axis is 
		reflective. Simulating reflectance properly requires 
		ray-tracing which cannot (yet) be done in real-time. We 
		are using the stencil buffer and suitable projections instead. 
		Only the first reflective geom in the model is rendered as 
		such. This adds one extra rendering pass through all geoms, 
		in addition to the extra rendering pass added by each 
		shadow-casting light.

		Returns
		-------
		float
		"""
		return self.asset.reflectance


	@reflectance.setter
	@blue.restrict
	def reflectance(self, reflectance: int|float) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.reflectance = reflectance


	@property
	def metallic(self) -> float:
		"""
		This attribute corresponds to uniform metallicity 
		coefficient applied to the entire material. This 
		attribute has no effect in MuJoCo’s native renderer, 
		but it can be useful when rendering scenes with a 
		physically-based renderer. In this case, if a 
		non-negative value is specified, this metallic value 
		should be multiplied by the metallic texture sampled 
		value to obtain the final metallicity of the material.

		Returns
		-------
		float
		"""
		return self.asset.metallic


	@metallic.setter
	@blue.restrict
	def metallic(self, metallic: int|float) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.metallic = metallic


	@property
	def roughness(self) -> float:
		"""
		This attribute corresponds to uniform roughness 
		coefficient applied to the entire material. This 
		attribute has no effect in MuJoCo’s native renderer, 
		but it can be useful when rendering scenes with a 
		physically-based renderer. In this case, if a non-negative 
		value is specified, this roughness value should be 
		multiplied by the roughness texture sampled value to 
		obtain the final roughness of the material.

		Returns
		-------
		float
		"""
		return self.asset.roughness


	@roughness.setter
	@blue.restrict
	def roughness(self, roughness: int|float) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.roughness = roughness


	@property
	def color(self) -> blue.ColorType:
		"""
		The color of the Material if set will be multiplied 
		by the Things Color and Texture.

		Returns
		-------
		blue.ThingType
		"""
		return self.asset.color


	@color.setter
	@blue.restrict
	def color(self, color: str|object|None) -> None:
		#print(self.ID, self.asset._references)
		#print(self in self.asset._references)
		self.asset._prepare_for_modification(self)
		self.asset.color = color


	@property
	def texture(self) -> blue.TextureType:
		"""
		Referencing the material from a Thing will cause the 
		texture to be applied to that Thing.

		Returns
		-------
		blue.TextureType
		"""
		return self.asset.texture


	@texture.setter
	@blue.restrict
	def texture(self, texture: blue.TextureType) -> None:
		self.asset._prepare_for_modification(self)
		self.asset.texture = texture#.copy()



"""
This package contains user utilities, structural subroutines and backend functions.

User Utilities
^^^^^^^^^^^^^^
- :mod:`Geometry <blueprints.utils.geometry>`
- :mod:`Perlin Noise <blueprints.utils.perlin>`

Structural Subroutines
^^^^^^^^^^^^^^^^^^^^^^
- :mod:`Views <blueprints.utils.view>`
- :mod:`Lattice <blueprints.utils.lattice>`

Backend Functions
^^^^^^^^^^^^^^^^^
- :mod:`Type Safty <blueprints.utils.typechecker>`
- :mod:`Naming <blueprints.utils.naming>`
- :mod:`Register <blueprints.utils.register>`
"""


"""
To make blueprints amneable for multi agent settings the :class:`Agent` class provides easily structured
access to the relevant Things for observation (:class:`Sensors <blueprints.sensors.BaseSensor>` and :class:`Cameras <blueprints.Camera>`) 
and action (:class:`Actuators <blueprints.actuators.General>`).

Agents are used similar to :class:`Bodies <blueprints.bodies.Body>` with some minor differences.
The child attributes for normal Things map only to direct children, but for an :class:`Agent` instance 
the attributes :attr:`actuators <Agent.actuators>`, :attr:`sensors <Agent.sensors>` and :attr:`cameras <Agent.cameras>` 
map to all descendants as a more intuitive shortcut.
All observations from Sensors and Cameras are bundeled in an :attr:`observation <Agent.observation>` attribute and 
all actions can be applied through the :attr:`force <Agent.force>` and :attr:`activation <Agent.activation>` attribute.

.. code-block:: mxml
	:caption: Agent XML structure
	
	<sensors>
		<force site='site_1'>
		<force site='site_2'>
	</sensors>
	<actuators>
		<general body='AGENT:body_1' dynprm='filter'>
		<general body='body_2'>
	</actuators>
	<body name='AGENT:body_1'>
		<body name='body_2'>
			<site name='site_1'>
		</body>
		<site name='site_2'>
		<camera name='cam'>
		<joint name='joint'>
	</body>

.. code-block::
	:caption: Agent Interface

	>>> agent.observation
	{'<Camera>cam': np.ndarray([...], shape=(width, height)), 
	 '<Force>_(0)': np.ndarray([0.]), 
	 '<Force>_(1)': np.ndarray([0.])}
	>>> agent.action_shape
	{'force': 1, 
	 'activation': 1}
	>>> agent.force = np.darray([...])
	>>> agent.activation = = np.darray([...])



Additionally shapes for observations and actions are available as attributes as well.
"""

import blueprints as blue
import numpy as np



class Agent(blue.AgentType, blue.Body):
	"""
	This class emulates a Body which serves as an interface for 
	sensors, cameras and actuators of an Agent. It provides uniform 
	access to the observation and action relevant Things.
	"""
	def __str__(self) -> str:
		"""
		Returns
		-------
		str
			A representation of the Agent. The xml will be parsed as :class:`Body <blue.bodies.Body>`.
		"""
		return f'<agent/body name="{self.name}">'


	@property
	def name(self) -> str:
		"""
		The returned name might differ from the user specification by an enumerations 
		scheme that is applied if two Things of the same type in a kinematic tree share 
		the same name.
		
		Returns
		-------
		str
			possibly extended name of the Thing
		"""
		if self._name_scope is not None:
			return f"AGENT:{self._name_scope.name(self).replace('AGENT:', '')}"
		else:
			return f"AGENT:{self._name.replace('AGENT:', '')}"


	@blue.restrict
	@classmethod
	def _from_xml_element(cls,
			      xml_element,
			      actuators: list = []) -> blue.ThingType:
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		init_args['copy'] = False
		# Strip AGENT: prefix — the name property re-adds it
		name = post_args.pop('name', None)
		if name is not None:
			name = name.replace('AGENT:', '')
		init_args['name'] = name
		agent = object.__new__(cls)
		agent.__init__(**init_args)
		for key, val in post_args.items():
			setattr(agent, key, val)
		for actuator in actuators:
			agent.attach(actuator, copy=False)
			actuator.body = agent
		return agent


	# AGENT PROPERTIES

	@property
	def cameras(self) -> blue.ViewType:
		"""
		Instead of returning just the children all descending Cameras are returned.
		
		Returns
		-------
		View
			The view is routed to all cameras of the agent (instead of just those that follow directly in the kinematic hierarchy).
		"""
		cameras = self.descendants['cameras']['descendants']
		return blue.View(cameras, 'cameras', self)


	@property
	def sensors(self) -> blue.ViewType:
		"""
		Instead of returning just the children all descending Sensors are returned.
		
		Returns
		-------
		View
			The view is routed to all sensors of the agent (instead of just those that follow directly in the kinematic hierarchy).
		"""
		sensors = self.descendants['sensors']['descendants']
		return blue.View(sensors, 'sensors', self)


	@property
	def actuators(self) -> blue.ViewType:
		"""
		Instead of returning just the children all descending Actuators are returned.
		
		Returns
		-------
		View
			The view is routed to all actuators of the agent (instead of just those that follow directly in the kinematic hierarchy).
		"""
		actuators = self.descendants['actuators']['descendants']
		return blue.View(actuators, 'actuators', self)


	@property
	def sensor_observation_shape(self) -> dict:
		"""
		The dictionary contains the names of all sensors and their shapes.
		
		Returns
		-------
		dict
		"""
		if not self.root._built:
			raise Exception('Agent.observation_shape is only accessable after the World has been built.')
		return {f'<{sensor.__class__.__name__}>{sensor.name}': (sensor.DIMENSIONS,) \
			for sensor in self.descendants['sensors']['descendants']}


	@property
	def camera_observation_shape(self) -> dict:
		"""
		The dictionary contains the names of all cameras and their shapes.
		
		Returns
		-------
		dict
		"""
		if not self.root._built:
			raise Exception('Agent.observation_shape is only accessable after the World has been built.')
		return {f'<{camera.__class__.__name__}>{camera.name}': tuple(camera.resolution) \
			for camera in self.descendants['cameras']['descendants']}


	@property
	def observation_shape(self) -> dict:
		"""
		The dictionary contains the names of all observables and their shapes.
		
		Returns
		-------
		dict
		"""
		if not self.root._built:
			raise Exception('Agent.observation_shape is only accessable after the World has been built.')
		cameras = self.camera_observation_shape
		sensors = self.sensor_observation_shape
		return {**cameras, **sensors}


	@property
	def sensor_observation(self) -> dict:
		"""
		The dictionary contains the name of all sensor observables and their data.
		
		Returns
		-------
		dict
		"""
		if not self.root._built:
			raise Exception('Agent.observation is only accessable after the World has been built.')
		return {f'<{sensor.__class__.__name__}>{sensor.name}': sensor.observation \
			for sensor in self.descendants['sensors']['descendants']}


	@property
	def camera_observation(self) -> dict:
		"""
		The dictionary contains the name of all camera observables and their data.
		
		Returns
		-------
		dict
		"""
		if not self.root._built:
			raise Exception('Agent.observation is only accessable after the World has been built.')
		return {f'<{camera.__class__.__name__}>{camera.name}': camera.observation \
			for camera in self.descendants['cameras']['descendants']}
		
	@property
	def observation(self) -> dict:
		"""
		The dictionary contains the name of all observables and their data.
		
		Returns
		-------
		dict
		"""
		if not self.root._built:
			raise Exception('Agent.observation is only accessable after the World has been built.')
		cameras = self.camera_observation
		sensors = self.sensor_observation
		return {**cameras, **sensors}


	@property
	def activation(self) -> list:
		"""
		All activations of Actuators that have an activation state. 
		Actuators with ``'dyntype' == 'none'`` do not have an activation.
		
		Returns
		-------
		list
		"""
		activation = self.actuators.activation
		return [x for x in activation if x is not None]


	@activation.setter
	@blue.restrict
	def activation(self, 
		       activation: list[int|float]|np.ndarray) -> None:
		"""
		A 1-D vector of activations for the actuators activation states.
		
		Parameters
		----------
		activation : list[int | float] | np.ndarray
		"""
		actuators = self.actuators
		mask = [x is not None for x in activation]
		for i, (x, m) in enumerate(zip(activation, mask)):
			if m:
				actuators[i] = x


	@property
	def force(self) -> list:
		"""
		All forces of Actuators of the Agent
		
		Returns
		-------
		list
		"""
		return self.actuators.force


	@force.setter
	@blue.restrict
	def force(self, 
		  force: list[int|float]|np.ndarray) -> None:
		"""
		Parameters
		----------
		force : list[int | float] | np.ndarray
			A 1-D vector of forces to be applied to the actuators
		"""
		actuators = self.actuators
		for i, f in enumerate(force):
			actuators[i].force = float(f)


	@property
	def action_shape(self) -> dict:
		"""
		The shapes of both action types (activations and forces)
		
		Returns
		-------
		dict
		"""
		if not self.root._built:
			raise Exception('Agent.action_shape is only accessable after the World has been built.')
		activation = self.activation
		force      = self.force
		return {'activation': len(activation), 'force': len(force)}



"""
The Blueprints library interfaces the Mujoco (Multi Joints Contact) physics simulation. 
To construct a :class:`World <blueprints.world.World>` you can combine many :mod:`Things <blueprints.thing.base>` 
via :meth:`attachment <blueprints.thing.node.NodeThing.attach>`, while blueprints translates 
your construction to xml in the background. Things are differentiated in concrete Things, which 
are the onse that you can use directly and abstract classes which implement a utilities for multiple 
concrete Things. Abstract Things are mostly located in :mod:`thing <blueprints.thing>` and should not be instantiated directly. 
Never the less they contain documentation relevant for many concrete Things, so you might want to 
have a look at it. Concrete Things are:

Concrete Things
---------------
- :mod:`Actuators <blueprints.actuators>` (All but ``BaseActuator``)
- :mod:`Bodies <blueprints.body>`
- :mod:`Camera <blueprints.camera>`
- :mod:`Geoms <blueprints.geoms>` (All but ``BaseGeom``)
- :mod:`Sites <blueprints.sites>` (All but ``BaseSite``)
- :mod:`Joints <blueprints.joints>` (All but ``BaseJoint``)
- :mod:`Light <blueprints.light>`
- :mod:`Materials <blueprints.material>`
- :mod:`Textures <blueprints.texture>` (All but ``BaseTexture``)
- :mod:`Sensors <blueprints.sensors>` (All but ``<Thing>Sensor``)
- :mod:`Tendons <blueprints.tendon>`  (Only ``Tendon``)
- :mod:`Worlds <blueprints.world.World>`

:mod:`Caches <blueprints.cache>` and :mod:`Assets <blueprints.assets>` are handeled in the background, so you don't 
have to create them manually and can most often treat their occurences in ordinary mujoco as 
another :meth:`attachment <blueprints.thing.node.NodeThing.attach>`.

Some helpful Things that are not part of the mujoco XML scheme but that are directly useable by 
the user are:

- :mod:`Agents <blueprints.agent>`
- :mod:`Lattices <blueprints.utils.lattice>`
- :mod:`Placeholders <blueprints.placeholder>`

Modificators
------------
Blueprints design is centered around the idea to 
reduce redundancy, that would be encountered in XML editing. Once you have created some Things 
you can easily produce on the fly modified copies through many helpful utility methods. We list 
some of the most important ones.


- :meth:`shift <blueprints.thing.moveable.MoveableThing.shift>`
- :meth:`locate <blueprints.thing.moveable.MoveableThing.locate>`
- :meth:`rotate <blueprints.thing.moveable.MoveableThing.rotate>`
- :meth:`align <blueprints.thing.moveable.MoveableThing.align>`
- :meth:`scaled <blueprints.thing.moveable.MoveableThing.scaled>`
- :meth:`Camera.looking <blueprints.camera.Camera.looking>`
- :meth:`copy <blueprints.thing.base.BaseThing.copy>`

The ``copy`` method in particular can be used to perform all sorts of modifications 
by passing the desired changes as keyword arguments.

Kinematic Tree
--------------
To modify the kinematic hierarchy there are some methods and attributes to have a look at:

- :attr:`parent <blueprints.thing.base.BaseThing.parent>` returns the parent of a Thing. If the Thing has no parent it returns ``None``.
- :attr:`root <blueprints.thing.base.BaseThing.root>` returns the root of a Things kinematic hierarchy (which might be itself).
- :attr:`path <blueprints.thing.base.BaseThing.path>` returns a list containing the path from a Thing to its root.
- :meth:`attach <blueprints.thing.node.NodeThing.attach>` attaches copies of any number of Things.
- :meth:`detach <blueprints.thing.node.NodeThing.detach>` reverses the effect of ``attach``.
- :meth:`all <blueprints.thing.node.NodeThing.all>` returns a :mod:`View <blueprints.utils.view>` of all children.
- :mod:`NodeThing.\\<children_type\\> <blueprints.utils.view>` returns a  :mod:`View <blueprints.utils.view>` of the given children type.

Alternatively kinematic trees can also be constructed by passing children as additional arguments to the ``__init__`` of 
all concrete :class:`NodeThings <blueprints.thing.node.NodeThing>`.

Types
-----

To enable typesafty :meth:`restrictions <blueprints.utils.typechecker.restrict>` via type hints, the 
inheritance hierarchy of :class:`Things <blueprints.thing.base.BaseThing>` is mirrored in :mod:`types <blueprints.mirrortypes>`.


.. inheritance-diagram:: blueprints.mirrortypes
    :parts: 1

"""


from collections import defaultdict
from functools import wraps


WARNING_FLAG = defaultdict(lambda *args: False)


def _experimental(func):
	@wraps(func)
	def wrapper(*args, **kwargs):
		if not WARNING_FLAG[func.__name__]:
			WARNING_FLAG[func.__name__] = True
			print(f'WARNING: {func.__qualname__} is an experimental feature likely to cause problems!')
		return func(*args, **kwargs)
	return wrapper


from .utils.typechecker import restrict

from .mirrortypes       import *
from .utils.view  import View, AllView, LatticeView

from .thing       import *

from .utils.lattice import Lattice
from .utils.perlin  import perlin
from .thing.focal import FocalThing

from .placeholder import Placeholder
from .body        import Body
from .tendon      import Tendon
from .world       import World

from . import assets
from . import tube
from . import sites
from . import geoms
from .utils import geometry
from . import sensors
from .utils import naming
from . import joints
from . import cache
from . import actuators

from .camera   import Camera
from .light    import Light
from .geoms    import BaseGeom
from .sites    import BaseSite
from .joints   import BaseJoint
from .material import Material
from .texture  import *
from .agent    import Agent

from .utils import register

#from .environment import Environment
from blueprints.utils.geometry import TAU, PI, DEGREES_TO_RADIANS, RADIANS_TO_DEGREES, Vector, Rotation

REGISTER = register.Register()


import numpy as np
import xml.etree.ElementTree as xml
import blueprints as blue



class BaseJoint(blue.JointType, blue.MoveableThing, blue.thing.NodeThing):

	"""
	The Joint class represents Mujoco joint objects. All Joints inherit from this base class.

	.. note::
		Joints enable movement along the degrees of freedom their parent Bodies and not to Bodies 
		attached to the Joints (one cannot attach Bodies to Joints). This might be counterintuitive 
		since real joints move objects which are attached to them.

	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-joint>`__.
	"""

	@blue.restrict
	def __init__(self, 
		     pos:                np.ndarray|list[int|float] = [0., 0., 0.], 
		     alpha:              int|float = 0., 
		     beta:               int|float = 0., 
		     gamma:              int|float = 0., 
		     springdamper:       np.ndarray|list[int|float] = [0., 0.], 
		     actuatorforcerange: np.ndarray|list[int|float] = [0., 0.], 
		     stiffness:          int|float = 0., 
		     springref:          int|float = 0., 
		     armature:           int|float = 0., 
		     damping:            int|float = 0., 
		     sensors:            blue.SensorType|list[blue.SensorType]|None     = None, 
		     actuators:          blue.ActuatorType|list[blue.ActuatorType]|None = None, 
		     name:               str|None  = None, 
		     copy:               bool      = True, 
		     x:                  int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:                  int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:                  int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : np.ndarray | list[int | float], optional
			Represents the position of the object. Changing this attribute also changes the properties 
			:attr:`x`, :attr:`y` and :attr:`z`.
		alpha : int | float, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also 
			changes the :attr:`euler` property.
		beta : int | float, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also 
			changes the :attr:`euler` property.
		gamma : int | float, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also 
			changes the :attr:`euler` property.
		springdamper : np.ndarray | list[int | float], optional
			When both numbers are positive, the compiler will override any stiffness and damping 
			values specified with the attributes below, and will instead set them automatically so 
			that the resulting mass-spring-damper for this joint has the desired time constant 
			(first value) and damping ratio (second value). 
		actuatorforcerange : np.ndarray | list[int | float], optional
			This attribute specifies whether actuator forces acting on the Joint should be clamped.
		stiffness : int | float, optional
			Joint stiffness. If this value is positive, a spring will be created with equilibrium 
			position given by springref below. The spring force is computed along with the other 
			passive forces.
		springref : int | float, optional
			The Joint position or angle in which the joint spring (if any) achieves equilibrium.
		armature : int | float, optional
			Armature inertia (or rotor inertia, or reflected inertia) of all degrees of freedom 
			created by this Joint. These are constants added to the diagonal of the inertia matrix 
			in generalized coordinates. They make the simulation more stable, and often increase 
			physical realism. This is because when a motor is attached to the system with a transmission 
			that amplifies the motor force by :math:`c`, the inertia of the rotor (i.e., the moving part 
			of the motor) is amplified by :math:`c\\cdot c`. The same holds for gears in the early 
			stages of planetary gear boxes. These extra inertias often dominate the inertias of the 
			robot parts that are represented explicitly in the model, and the armature attribute is the 
			way to model them.
		damping : int | float, optional
			Damping applied to all degrees of freedom created by this joint. Unlike friction loss which 
			is computed by the constraint solver, damping is simply a force linear in velocity. It is 
			included in the passive forces. Despite this simplicity, larger damping values can make 
			numerical integrators unstable, which is why our Euler integrator handles damping implicitly.
		sensors : blue.SensorType | list[blue.SensorType] | None, optional
			The Sensors are attached to the actuator on initialization.
		actuators : blue.ActuatorType | list[blue.ActuatorType] | None, optional
			The Actuators are attached to the actuator on initialization.
		name : str | None, optional
			The user specified name. In the case of a naming conflict the name will be altered by an 
			enumeration scheme.
		copy : bool, optional
			If True, the Sensors and Actuators passed during init will be copied before attachment, 
			otherwise the original Sensor is attached.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		pack      = lambda x: x if isinstance(x, list) else [x]
		sensors   = pack(sensors)   if sensors   else []
		actuators = pack(actuators) if actuators else []
		self._sensors           = []
		self._tendons           = []
		self._actuators         = []
		self._CHILDREN          = {'tendons':   {'type':     blue.TendonType, 
							 'children': self._tendons}, 
					   'sensors':   {'type':     blue.SensorType, 
							 'children': self._sensors}, 
					   'actuators': {'type':     blue.ActuatorType, 
							 'children': self._actuators}}
		self.springdamper       = springdamper
		self.actuatorforcerange = actuatorforcerange
		self.stiffness          = stiffness
		self.springref          = springref
		self.armature           = armature
		self.damping            = damping
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 **kwargs)
		self.attach(*sensors, 
			    *actuators, 
			    copy=copy)
		for actuator in self.actuators:
			actuator.joint = self
		self._check_children_types()


	@blue.restrict
	def _build(self, 
		   parent: xml.Element, 
		   world, 
		   indicies, 
		   **kwargs) -> xml.Element:
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Joint.
		"""
		self._index      = indicies['jnt']
		indicies['jnt'] += 1
		if hasattr(self, 'axis'):
			if 'axis' in kwargs:
				axis = kwargs.pop('axis')
			else:
				axis = self.axis
			kwargs['axis'] = self.rotation_matrix @ axis
		return super()._build(parent, world, indicies, **kwargs)


	@blue.restrict
	@classmethod
	def _from_xml_element(cls, 
				  xml_element: xml.Element, 
				  sensors:     list = None, 
				  actuators:   list = []) -> blue.ThingType:
		"""
		This method reconstructs a Joint from an xml element.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a Geom is reconstructed.
		sensors : blue.SensorType | list[blue.SensorType] | None, optional
			The Sensors are attached to the actuator on initialization.
		actuators : blue.ActuatorType | list[blue.ActuatorType] | None, optional
			The Actuators are attached to the actuator on initialization.
		
		Returns
		-------
		blue.ThingType
			The reconstructed Joint.
		"""
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		init_args['sensors'] = sensors
		init_args['copy']    = False
		if 'type' in rest_args:
			joint_type = rest_args.pop('type')
		else:
			joint_type = None
		joint = object.__new__(blue.REGISTER.JOINT_THINGS[joint_type])
		joint.__init__(**init_args)
		for key, val in post_args.items():
			setattr(joint, key, val)
		for actuator in actuators:
			joint.attach(actuator, copy=False)
			actuator.joint = joint
		return joint

	# MUJOCO PROPERTIES

	@property
	def type(self) -> str:
		"""
		This is a derived attribute, which is used to specify the ``type`` attribute in the 
		mujoco ``joint`` tag.

		Returns
		-------
		str
			The type is the lower case name of the Joint class.
		"""
		return self.__class__.__name__.lower()


	@property
	def springdamper(self) -> np.ndarray:
		"""
		When both numbers are positive, the compiler will override any stiffness and damping 
		values specified with the attributes below, and will instead set them automatically so 
		that the resulting mass-spring-damper for this joint has the desired time constant 
		(first value) and damping ratio (second value). 
		
		Returns
		-------
		np.ndarray
		"""
		return self._springdamper


	@springdamper.setter
	@blue.restrict
	def springdamper(self, springdamper: np.ndarray|list[int|float]) -> None:
		"""
		When both numbers are positive, the compiler will override any stiffness and damping 
		values specified with the attributes below, and will instead set them automatically so 
		that the resulting mass-spring-damper for this joint has the desired time constant 
		(first value) and damping ratio (second value). 
		
		Parameters
		----------
		springdamper : np.ndarray | list[int | float]
			The springdamer which is assigned.
		"""
		if isinstance(springdamper, list):
			if len(springdamper) != 2:
				raise ValueError(f'springdamper must have a length of 2, got {len(springdamper)} instead.')
		elif isinstance(springdamper, np.ndarray):
			if springdamper.shape != (2,):
				raise ValueError(f'springdamper must have a shape of (2,), got {springdamper.shape} instead.')
		self._springdamper = np.array(springdamper, dtype=np.float32)


	@property
	def actuatorforcerange(self) -> np.ndarray:
		"""
		actuatorforcerange : np.ndarray | list[int | float], optional
		This attribute specifies whether actuator forces acting on the Joint should be clamped.
		
		Returns
		-------
		np.ndarray
		"""
		return self._actuatorforcerange


	@actuatorforcerange.setter
	@blue.restrict
	def actuatorforcerange(self, actuatorforcerange: np.ndarray|list[int|float]) -> None:
		"""
		This attribute specifies whether actuator forces acting on the Joint should be clamped.
		
		Parameters
		----------
		actuatorforcerange : np.ndarray | list[int | float], optional
			The actuator force range which is assigned.
		"""
		if isinstance(actuatorforcerange, list):
			if len(actuatorforcerange) != 2:
				raise ValueError(f'actuatorforcerange must have a length of 2, got {len(actuatorforcerange)} instead.')
		elif isinstance(actuatorforcerange, np.ndarray):
			if actuatorforcerange.shape != (2,):
				raise ValueError(f'actuatorforcerange must have a shape of (2,), got {actuatorforcerange.shape} instead.')
		self._actuatorforcerange = np.array(actuatorforcerange, dtype=np.float32)


	@property
	def stiffness(self) -> float:
		"""
		Joint stiffness. If this value is positive, a spring will be created with equilibrium 
		position given by springref below. The spring force is computed along with the other 
		passive forces.
		
		Returns
		-------
		float
		"""
		return self._stiffness


	@stiffness.setter
	@blue.restrict
	def stiffness(self, stiffness: int|float) -> None:
		"""
		Joint stiffness. If this value is positive, a spring will be created with equilibrium 
		position given by springref below. The spring force is computed along with the other 
		passive forces.
		
		Parameters
		----------
		stiffness : int | float, optional
			The stiffness which is assigned
		"""
		self._stiffness = float(stiffness)


	@property
	def springref(self) -> float:
		"""
		The Joint position or angle in which the joint spring (if any) achieves equilibrium.
		
		Returns
		-------
		float
		"""
		return self._springref


	@springref.setter
	@blue.restrict
	def springref(self, springref: int|float) -> None:
		"""
		The Joint position or angle in which the joint spring (if any) achieves equilibrium.
		
		Parameters
		----------
		springref : int | float, optional
			The springref which is assigned
		"""
		self._springref = float(springref)


	@property
	def armature(self) -> float:
		"""
		Armature inertia (or rotor inertia, or reflected inertia) of all degrees of freedom 
		created by this Joint.
		
		Returns
		-------
		float
		"""
		return self._armature


	@armature.setter
	@blue.restrict
	def armature(self, armature: int|float) -> None:
		"""
		Armature inertia (or rotor inertia, or reflected inertia) of all degrees of freedom created 
		by this Joint.
		
		Parameters
		----------
		armature : int | float
			The armature which is assigned
		"""
		self._armature = float(armature)


	@property
	def damping(self) -> float:
		"""
		Damping applied to all degrees of freedom created by this joint. Unlike friction loss which 
		is computed by the constraint solver, damping is simply a force linear in velocity. It is 
		included in the passive forces. Despite this simplicity, larger damping values can make 
		numerical integrators unstable, which is why our Euler integrator handles damping implicitly.
		
		Returns
		-------
		float
		"""
		return self._damping


	@damping.setter
	@blue.restrict
	def damping(self, damping: int|float) -> None:
		"""
		Damping applied to all degrees of freedom created by this joint. Unlike friction loss which is 
		computed by the constraint solver, damping is simply a force linear in velocity. It is included 
		in the passive forces. Despite this simplicity, larger damping values can make numerical 
		integrators unstable, which is why our Euler integrator handles damping implicitly.
		
		Parameters
		----------
		damping : int | float
			The damping which is assigned
		"""
		self._damping = float(damping)



class Hinge(blue.HingeType, BaseJoint):

	"""
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-joint>`__.
	"""

	@blue.restrict
	def __init__(self, 
		     pos:          np.ndarray|list[int|float]     = [0., 0., 0.], 
		     axis:         np.ndarray|list[int|float]|str = [0., 0., 1.], 
		     range:        np.ndarray|list[int|float]     = [0., 0.], 
		     ref:          int|float = 0., 
		     frictionloss: int|float = 0., 
		     name:         str|None  = None, 
		     x:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : np.ndarray | list[int | float], optional
			Represents the position of the object. Changing this attribute also changes the properties 
			:attr:`x`, :attr:`y` and :attr:`z`.
		axis : np.ndarray | list[int | float] | str, optional
			This attribute specifies the axis of rotation for hinge Joints and the direction of 
			translation for Slides.
		range : np.ndarray | list[int | float], optional
			The reference position or angle of the Joint. It defines the joint value corresponding 
			to the initial model configuration.
		ref : int | float, optional
			The reference position or angle of the Joint. It defines the joint value corresponding 
			to the initial model configuration.
		frictionloss : int | float, optional
			Friction loss due to dry friction. This value is the same for all degrees of freedom 
			created by this Joint.
		name : str | None, optional
			The user specified name. In the case of a naming conflict the name will be altered by 
			an enumeration scheme.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		axis = blue.utils.geometry.Vector.get_axis(axis)
		self.axis         = axis
		self.range        = range
		self.ref          = ref
		self.frictionloss = frictionloss
		super().__init__(pos=pos, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 **kwargs)


	@property
	def axis(self) -> np.ndarray:
		"""
		This attribute specifies the axis of rotation for hinge Joints and the direction of translation for Slides.
		
		Returns
		-------
		np.ndarray
		"""
		return self._axis


	@axis.setter
	@blue.restrict
	def axis(self, axis: np.ndarray|list[int|float]) -> None:
		"""
		This attribute specifies the axis of rotation for hinge Joints and the direction of translation for Slides.
		
		Parameters
		----------
		axis : np.ndarray | list[int | float]
			The axis which is assigned
		"""
		self._axis = np.array(axis, np.float32)


	@property
	def range(self) -> np.ndarray:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Returns
		-------
		np.ndarray
		"""
		return self._range


	@range.setter
	@blue.restrict
	def range(self, range: np.ndarray|list[int|float]) -> None:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Parameters
		----------
		range : np.ndarray | list[int | float]
			The range which is assigned
		"""
		self._range = np.array(range, np.float32)


	@property
	def ref(self) -> float:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Returns
		-------
		float
		"""
		return self._ref


	@ref.setter
	@blue.restrict
	def ref(self, ref: int|float) -> None:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Parameters
		----------
		ref : int | float
			The ref which is assigned
		"""
		self._ref = float(ref)


	@property
	def frictionloss(self) -> float:
		"""
		Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		
		Returns
		-------
		float
		"""
		return self._frictionloss


	@frictionloss.setter
	@blue.restrict
	def frictionloss(self, frictionloss: int|float) -> None:
		"""
		Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		
		Parameters
		----------
		frictionloss : int | float
			The xxx wfrictionlossh is assigned
		"""
		self._frictionloss = float(frictionloss)



class Slide(blue.SlideType, BaseJoint):

	"""
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-joint>`__.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:          np.ndarray|list[int|float]     = [0., 0., 0.], 
		     axis:         np.ndarray|list[int|float]|str = [0., 0., 1.], 
		     range:        np.ndarray|list[int|float]     = [0., 0.], 
		     ref:          int|float = 0, 
		     frictionloss: int|float = 0., 
		     name:         str|None  = None, 
		     x:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : np.ndarray | list[int | float], optional
			Represents the position of the object. Changing this attribute also changes the properties 
			:attr:`x`, :attr:`y` and :attr:`z`.
		axis : np.ndarray | list[int | float] | str, optional
			This attribute specifies the axis of rotation for hinge Joints and the direction of translation for Slides.
		range : np.ndarray | list[int | float], optional
			The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		ref : int | float, optional
			The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		frictionloss : int | float, optional
			Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		name : str | None, optional
			The user specified name. In the case of a naming conflict the name will be altered by an 
			enumeration scheme.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		axis = blue.utils.geometry.Vector.get_axis(axis)
		self.axis         = axis
		self.range        = range
		self.ref          = ref
		self.frictionloss = frictionloss
		super().__init__(pos=pos, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 **kwargs)


	@property
	def axis(self) -> np.ndarray:
		"""
		This attribute specifies the axis of rotation for hinge Joints and the direction of translation for Slides.
		
		Returns
		-------
		np.ndarray
		"""
		return self._axis


	@axis.setter
	@blue.restrict
	def axis(self, axis: np.ndarray|list[int|float]) -> None:
		"""
		This attribute specifies the axis of rotation for hinge Joints and the direction of translation for Slides.
		
		Parameters
		----------
		axis : np.ndarray | list[int | float]
			The axis which is assigned
		"""
		self._axis = np.array(axis, np.float32)


	@property
	def range(self) -> np.ndarray:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Returns
		-------
		np.ndarray
		"""
		return self._range


	@range.setter
	@blue.restrict
	def range(self, range: np.ndarray|list[int|float]) -> None:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Parameters
		----------
		range : np.ndarray | list[int | float]
			The range which is assigned
		"""
		self._range = np.array(range, np.float32)


	@property
	def ref(self) -> float:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Returns
		-------
		float
		"""
		return self._ref


	@ref.setter
	@blue.restrict
	def ref(self, ref: int|float) -> None:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Parameters
		----------
		ref : int | float
			The ref which is assigned
		"""
		self._ref = float(ref)


	@property
	def frictionloss(self) -> float:
		"""
		Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		
		Returns
		-------
		float
		"""
		return self._frictionloss


	@frictionloss.setter
	@blue.restrict
	def frictionloss(self, frictionloss: int|float) -> None:
		"""
		Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		
		Parameters
		----------
		frictionloss : int | float
			The xxx wfrictionlossh is assigned
		"""
		self._frictionloss = float(frictionloss)



class Ball(blue.BallType, BaseJoint):

	"""
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-joint>`__.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:          np.ndarray|list[int|float] = [0., 0., 0.], 
		     range:        np.ndarray|list[int|float] = [0., 0.], 
		     frictionloss: int|float                  = 0., 
		     name:         str|None                   = None, 
		     x:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:            int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : np.ndarray | list[int | float], optional
			Represents the position of the object. Changing this attribute also changes the properties 
			:attr:`x`, :attr:`y` and :attr:`z`.
		range : np.ndarray | list[int | float], optional
			The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		frictionloss : int | float, optional
			Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		name : str | None, optional
			The user specified name. In the case of a naming conflict the name will be altered by an 
			enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.range        = range
		self.frictionloss = frictionloss
		super().__init__(pos=pos, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 **kwargs)


	@property
	def range(self) -> np.ndarray:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Returns
		-------
		np.ndarray
		"""
		return self._range


	@range.setter
	@blue.restrict
	def range(self, range: np.ndarray|list[int|float]) -> None:
		"""
		The reference position or angle of the Joint. It defines the joint value corresponding to the initial model configuration.
		
		Parameters
		----------
		range : np.ndarray | list[int | float]
			The range which is assigned
		"""
		self._range = np.array(range, np.float32)


	@property
	def frictionloss(self) -> float:
		"""
		Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		
		Returns
		-------
		float
		"""
		return self._frictionloss


	@frictionloss.setter
	@blue.restrict
	def frictionloss(self, frictionloss: int|float) -> None:
		"""
		Friction loss due to dry friction. This value is the same for all degrees of freedom created by this Joint.
		
		Parameters
		----------
		frictionloss : int | float
			The xxx which is assigned
		"""
		self._frictionloss = float(frictionloss)



class Free(blue.FreeType, BaseJoint):

	"""
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-joint>`__.
	"""
	
	def __init__(self, 
		     **kwargs):
		"""
		Parameters
		----------
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		super().__init__(**kwargs)



class Joint(Hinge):

	"""
	Dummy for :class:`Hinge`
	"""

	@property
	def type(self) -> None:
		"""
		Dummy for :class:`Hinge`
		
		Returns
		-------
		str
			``'hinge'``
		"""
		return None



JOINT_THINGS = {None:   Joint, 
	       'hinge': Hinge, 
	       'slide': Slide,
	       'ball':  Ball, 
	       'free':  Free}
